/*  Student Name: Wade Tregaskis
    Student Number: 02557793 */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "ioutils.h"


/* displayRepeat(ostream& out, const char ch, int times)
     Outputs to the stream 'out' a sequence of identical characters - 'char ch'.
     It does so to the tune of 'times'.
     Note: The use of the 'times' parameter instead of a local variable is not
     erroneous. */

void displayRepeat(ostream& out, const char ch, int times) {
    while (times > 0) {
        out << ch;
        --times;
    }
}


/* outputFixedLength(ostream& out, const char *data, const int maxLength, const bool left)
     Outputs the cstring 'data' to 'out', establishing an output of fixed length,
     set by 'maxLength', and aligned to left or right, depending on 'left'.  Most
     useful for the fact that if 'data' is too long it will be truncated and an
     elipse automagically appended to indicate this. */

void outputFixedLength(ostream& out, const char *data, const int maxLength, const bool left) {
    char *result;
    int i;

    if (left) {
        cout.setf(ios::left);
    } else {
        cout.setf(ios::right);
    }

    cout.width(maxLength);

    if (strlen(data) > maxLength) { // Need to truncate it
        result = new char[maxLength + 1];
        strncpy(result, data, maxLength - 3);

        for (i = maxLength - 3; i < maxLength; ++i) {
            result[i] = '.';
        }

        result[maxLength] = '\0';

        cout << result;

        delete[] result;
    } else { // Fine as it is
        cout << data;
    }
}

/* safeGetline(istream &in, char *data, const int maxLength)
     Same general functionality as getline in terms of input retrieval, except
     for two things.  First, in the case where there is too much data it clears
     the error flag and buffer [up to the next line] so "in" may be used again.
     Second, it eliminates all whitespace at the start and end of the input
     string. */

void safeGetline(istream& in, char *data, const int maxLength) {
    char *temp;

    if ((maxLength < 2) || in.fail()) {
        return;
    }

    temp = new char[maxLength];
    in.getline(temp, maxLength - 1);

    if (in.fail() && strlen(temp)) {
        in.clear();
        in.ignore(10000, '\n');
    }

    removeOutsideWhitespace(&temp);
    strncpy(data, temp, strlen(temp) + 1);
    delete[] temp;
}

/* getBoolean(istream& in)
     Retrieves a yes or no answer from the user.  Doesn't display any prompt
     initially, but will advise the user if they enter something invalid. */

bool getBoolean(istream& in) {
    char temp[LINE_BUFFER_SIZE];

    if (in.fail()) { // Stream broken
        return false;
    }

    while (1) {
        safeGetline(cin, temp, LINE_BUFFER_SIZE);

        if (isBoolean(temp)) {
            return boolFromString(temp);
        } else {
            cout << "Please enter a yes or no answer, e.g. 'yes' or 'no':  ";
        }
    }
}

/* getChar(istream& in)
     Returns a character input from 'in', and clears the input stream until the
     next new line. */

char getChar(istream& in) {
    char temp;

    if (in.fail()) { // Stream broken
        return '\0';
    }

    in >> temp;
    in.ignore(10000, '\n');

    return temp;
}

/* switchIsInArguments(const int argc, const char **argv, const char sw)
     Returns true if the switch 'sw' appears anywhere in the standard argument
     list 'argv'.  Allows for any number of switches per field, e.g. both "-p"
     and "-cFp" will return true for 'p'.  Case sensitive. */

bool switchIsInArguments(const int argc, const char **argv, const char sw) {
    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] == '-') { // switch
            for (int n = 1; n < strlen(argv[i]); ++n) {
                if (argv[i][n] == sw) {
                    return true;
                }
            }
        }
    }

    return false;
}

/* firstDataArgument(const int argc, const char **argv)
     Returns the index of the first data argument - i.e. the first argument which
     doesn't start with a '-'.  In the event that none is found, returns 0. */

int firstDataArgument(const int argc, const char **argv) {
    for (int i = 1; i < argc; ++i) {
        if (argv[i][0] != '-') { // switch
            return i;
        }
    }

    return 0;
}

bool fileIsReadable(const char* name) {
    ifstream fin;

    fin.open(name);
    
    if (fin.fail()) {
        return false;
    } else {
        fin.close();
        return true;
    }
}

// ---------------------
// IMPLEMENTATION:  Menu
// ---------------------

Menu::Menu(const char *newTitle) {
    title = newTitle;
    selection = -1;
    numberOfItems = 0;
}

void Menu::setTitle(const char *newTitle) {
    title = newTitle;
}

const char* Menu::getTitle(void) const {
    return title;
}

void Menu::append(const char *newItem, int key) {
    if ((newItem) && (numberOfItems < MENU_ITEM_LIMIT)) {
        item[numberOfItems] = newItem;
        itemKey[numberOfItems] = key;
        ++numberOfItems;
    }
}

void Menu::appendBreak(void) {
    if (numberOfItems < MENU_ITEM_LIMIT) {
        item[numberOfItems] = NULL;
        ++numberOfItems;
    }
}

int Menu::getSelection(void) const {
    return selection;
}

void Menu::setSelection(int index) {
    if ((index > -1) && (index < numberOfItems) && (item[index] != NULL)) {
        selection = index;
    }
}

void Menu::clearSelection(void) {
    selection = -1;
}

const char* Menu::selectedName(void) const {
    if (selection > -1) {
        return item[selection];
    } else {
        return NULL;
    }
}

int Menu::selectedKey(void) const {
    if (selection > -1) {
        return itemKey[selection];
    } else {
        return -1;
    }
}

bool Menu::isFull(void) const {
    return (numberOfItems == MENU_ITEM_LIMIT);
}

void Menu::displayNumeric(void) {
    int *keyList, realCount = 0;
    char choice = '\0';

    if (numberOfItems > 0) {
        keyList = new int[numberOfItems];

        if (title) {
            cout << "--<== " << title << " ==>--\n\n";
        }
        
        for(int i = 0; i < numberOfItems; ++i) {
            if (item[i]) {
                keyList[realCount] = i;
                cout << "  " << realCount + 1 << ". " << item[i] << endl;
                ++realCount;
            } else {
                cout << endl;
            }
        }

        cout << endl;

        while (1) {
            cout << "Selection: ";
            choice = getChar(cin);

            if ((choice < '1') || ((int)choice > (realCount + '0'))) {
                cout << "Invalid selection - please enter a number between 1 and " << realCount << " (inclusive).\n";
            } else {
                break;
            }
        }

        selection = keyList[choice - '1'];
        delete[] keyList;
    }    
}

void Menu::displayAlphabetic(void) {
    int *keyList, realCount = 0;
    char choice = '\0';

    if (numberOfItems) {
        keyList = new int[numberOfItems];

        if (title) {
            cout << "--<== " << title << " ==>--\n\n";
        }

        for(int i = 0; i < numberOfItems; ++i) {
            if (item[i]) {
                keyList[realCount] = i;
                cout << "  " << (char)('a' + realCount) << ". " << item[i] << endl;
                ++realCount;
            } else {
                cout << endl;
            }
        }

        cout << endl;

        while (1) {
            cout << "Selection: ";
            choice = tolower(getChar(cin));

            if ((choice < 'a') || (choice > ('a' + realCount - 1))) {
                cout << "Invalid selection - please enter a letter between 'a' and '" << (char)('a' + realCount - 1) << "' (inclusive).\n";
            } else {
                break;
            }
        }

        selection = keyList[choice - 'a'];
        delete[] keyList;
    }
}
