/*  Student Name: Wade Tregaskis
    Student Number: 02557793 */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "subject.h"


/* int main(int argc, const char **argv)
    Handles the main event loop.  Accepts parameters, including filename, from the
    shell, or gets them from the user if not present  Loads the inital database,
    runs the main logical loop including main menu, and stores data back to disk. */

int main(int argc, const char **argv) {
    char *fileName = NULL;
    int fileArgIndex;
    SubjectList *defaultList = NULL;
    Menu *mainMenu = NULL;

    srand(time(NULL));
    
#ifndef __APPLE__
    mtrace(); // Generates malloc trace's - oooh so useful :)
#endif
    
    if (switchIsInArguments(argc, argv, 'v')) {
        cout << "Subject version " << SUBJECT_VERSION << " (library v" << SUBJECTLIB_VERSION << ")\n";
        cout << "   Written by Wade Tregaskis.  Free to distribute.\n";
        cout << "   Produced under hypothetical license for the Terran Central Exploration Bureau.\n";

        return 0;
    }
    
    if (switchIsInArguments(argc, argv, 'h')) {
        cout << "Usage:       " << argv[0] << " [-hv] [file]\n";
        cout << "Parameters:\n     -h      Display this help\n"
                          << "     -v      Display version number [and quit]\n"
                          << "     file    Data file containing subject information\n";
        return 0;
    }
    
    fileArgIndex = firstDataArgument(argc, argv);
    
    if (fileArgIndex > 0) {
        fileName = copyOfString(argv[fileArgIndex]);
    } else {
        cout << "Name of file from which to load data (default is 'subject.dat'):  ";
        
        fileName = new char[FILENAME_BUFFER_SIZE];
        safeGetline(cin, fileName, FILENAME_BUFFER_SIZE);

        if (strlen(fileName) == 0) {
            delete[] fileName;
            fileName = copyOfString("subject.dat");
        }
    }

    defaultList = new SubjectList(fileName);

    if (!fileIsReadable(fileName)) {
        cout << "\nError: Unable to read subject from \"" << fileName << "\" - "
             << "it may be missing or in use.\nIf you continue, a new [empty] "
             << "database will be created.\n\nDo you wish to continue?  ";

        if (!getBoolean(cin)) {
            delete defaultList;
            return 0;
        }
    } else if (defaultList->count() == 0) {
        cout << "\nWarning: \"" << fileName << "\" contains no entries.\n";
    }

    cout << endl;
    
    while (1) {
        if (mainMenu) {
            delete mainMenu;
        }
        
        mainMenu = new Menu("TCEB Subject Classification Program");

        if (defaultList->count() > 0) {
            mainMenu->append("Browse Subject List", 1);
            mainMenu->append("Searching for Existing Subjects", 2);
        }

        mainMenu->append("Add New Subject", 3);
        mainMenu->appendBreak();
        mainMenu->append("Exit", 4);

        mainMenu->displayNumeric();    
        cout << endl;

        switch (mainMenu->selectedKey()) {
        case 1:
            displayList(defaultList); break;
        case 2:
            searchList(defaultList); break;
        case 3:
            addSubjectToList(defaultList); break;
        case 4:
            defaultList->saveToFile();

            delete mainMenu;
            delete defaultList;
            
            return 0;
        }
    }
}

/* displayList(SubjectList *list, bool format)
     Presents a list of all entries in 'list' to the user.  Note: It does not
     sort the database or interpret the results in any other way than needed to
     display them.
       Automatically asks the user for long or short form of list. */

void displayList(SubjectList *list) {
    char choice = '\0';

    if (list->count() > 0) {
        cout << "Long or short list format (l/s)? ";
        choice = getChar(cin);
    }

    cout << endl;
    
    displayList(list, (toupper(choice) == 'L'));
}

/* displayList(SubjectList *list, bool format)
     Presents a list of all entries in 'list' to the user.  Note: It does not
     sort the database or interpret the results in any other way than needed to
     display them.
       If 'format' is true, entries are displayed in long form, otherwise only
     in brief. */

void displayList(SubjectList *list, bool format) {
    Cursor<Subject> *subjectCursor = NULL;
    Cursor<Property> *propertyCursor = NULL;
    Subject *currentSubject = NULL;
    Property *currentProperty = NULL;
    const char *holder = NULL;
    
    if (list->count() == 0) {
        holder = list->sourceFile();

        if (holder) {
            cout << "There are no entries in the list \"" << holder << "\".\n";
        } else {
            cout << "There are no entries to list.\n";
        }
    } else {
        subjectCursor = new Cursor<Subject>(list);

        holder = list->sourceFile();
        
        if (holder) {
            cout << "The list from \"" << holder << "\" contains the following subject:\n\n";
        }

        if (format) {
            cout << "NAME      <PROPERTIES>\n\n";
        } else {
            cout << "NAME\n\n";
        }
        
        while ((currentSubject = subjectCursor->getCurrent()) != NULL) {
            holder = currentSubject->getName();

            if (holder) {
                cout << holder << endl;

                if (format) {
                    propertyCursor = new Cursor<Property>(currentSubject->propertyList());

                    while ((currentProperty = propertyCursor->getCurrent()) != NULL) {
                        displayRepeat(cout, ' ', 10);

                        if (currentProperty->isBoolean()) {
                            cout << currentProperty->getName() << endl;
                        } else {
                            cout << currentProperty->getName() << ": " << currentProperty->getValue() << endl;
                        }

                        propertyCursor->moveToNext();
                    }

                    delete propertyCursor;
                }
            }

            subjectCursor->moveToNext();
        }

        delete subjectCursor;

        cout << endl;
    }
}

/* addSubjectToList(SubjectList *list)
     Inputs the title, artist and length of a new SubjectItem, appending it to
     'list'. */

void addSubjectToList(SubjectList *list) {
    char input[LINE_BUFFER_SIZE];
    Subject *result = NULL;
    List<Property> *properties = NULL;
    Property *curProp = NULL;
    const char *holder = NULL;

    holder = list->sourceFile();

    if (holder) {
        cout << "Adding new subject to \"" << holder << "\".\n\n";
    }

    cout << "Subject name: ";
    safeGetline(cin, input, LINE_BUFFER_SIZE);

    if (strlen(input) == 0) {
        cout << "Treating lack of input as cancel - returning to menu.\n";
        return;
    }

    result = new Subject(copyOfString(input));
    properties = result->propertyList();

    cout << "\nNotes:  (a) When finished adding properties, enter a blank line for Name.\n";
    cout << "        (b) If a property is boolean, enter a blank line for Value.\n";

    while (1) {
        cout << "Property Name: ";
        safeGetline(cin, input, LINE_BUFFER_SIZE);

        if (strlen(input)) {
            curProp = new Property(copyOfString(input), NULL);
            properties->append(curProp);
            curProp->release();
            
            cout << "        Value: ";
            safeGetline(cin, input, LINE_BUFFER_SIZE);

            if (strlen(input)) {
                curProp->setValue(copyOfString(input));
            }
        } else {
            break;
        }
    }

    list->append(result);
    result->release();

    if (holder) {
        cout << "\nNew subject \"" << result->getName() << "\" added to \"" << holder << "\".\n\n";
    } else {
        cout << "\nNew subject \"" << result->getName() << "\" added.\n\n";
    }
}

void addSubjectToList(SubjectList *list, List<Question> *questions) {
    char input[LINE_BUFFER_SIZE];
    Subject *result = NULL;
    List<Property> *properties = NULL;
    Property *currentProperty = NULL;
    Question *currentQuestion = NULL;
    Cursor<Question> *questionCursor = NULL;
    Cursor<Property> *propertyCursor = NULL;
    const char *holder = NULL;

    cout << "New subject's name: ";
    safeGetline(cin, input, LINE_BUFFER_SIZE);

    if (strlen(input) == 0) {
        cout << "Treating lack of input as cancel - returning to menu.\n";
        return;
    }
    
    result = new Subject(copyOfString(input));
    properties = result->propertyList();

    questionCursor = new Cursor<Question>(questions);

    while ((currentQuestion = questionCursor->getCurrent()) != NULL) {
        if (currentQuestion->isBoolean()) {
            if (currentQuestion->getBooleanAnswer()) {
                currentProperty = new Property(copyOfString(currentQuestion->getQuestion()), NULL);
                properties->append(currentProperty);
                currentProperty->release();
            }
        } else {
            holder = currentQuestion->getStringAnswer();

            if (holder && (caseInsensitiveStringCompare(holder, "remove") != 0)) {
                currentProperty = new Property(copyOfString(currentQuestion->getQuestion()), copyOfString(currentQuestion->getStringAnswer()));
                properties->append(currentProperty);
                currentProperty->release();
            }
        }

        questionCursor->moveToNext();
    }

    delete questionCursor;
    
    if (properties->count()) {
        cout << "\nCurrent properties are as follows:\n\n";
        
        propertyCursor = new Cursor<Property>(properties);

        while ((currentProperty = propertyCursor->getCurrent()) != NULL) {
            if (currentProperty->isBoolean()) {
                cout << currentProperty->getName() << endl;
            } else {
                cout << currentProperty->getName() << ": " << currentProperty->getValue() << endl;
            }

            propertyCursor->moveToNext();
        }

        delete propertyCursor;
    }

    cout << "\nNotes:  (a) When finished adding additional properties, enter a blank line for Name.\n";
    cout << "        (b) If a property is boolean, enter a blank line for Value.\n";

    while (1) {
        cout << "Property Name: ";
        safeGetline(cin, input, LINE_BUFFER_SIZE);

        if (strlen(input)) {
            currentProperty = new Property(copyOfString(input), NULL);
            properties->append(currentProperty);
            currentProperty->release();

            cout << "        Value: ";
            safeGetline(cin, input, LINE_BUFFER_SIZE);

            if (strlen(input)) {
                currentProperty->setValue(copyOfString(input));
            }
        } else {
            break;
        }
    }

    list->append(result);
    result->release();

    holder = list->sourceFile();
    
    if (holder) {
        cout << "\nNew subject \"" << result->getName() << "\" added to \"" << holder << "\".\n\n";
    } else {
        cout << "\nNew subject \"" << result->getName() << "\" added.\n\n";
    }
}

/* searchList(SubjectList *list)
     Uses repeatitive user input to search for a particular subject, showing an
     increasingly narrow list of options until the final singular item is found,
     or no match is found, in which case the option to add a new entry is made
     available.
       Dev Note:  This code is bad.  It just screams "I am evil!  Recode me!".
     Unfortunately this is not easily possible with the current class-less dev
     model. */

void searchList(SubjectList *list) {
    int lastCount;
    List<Question> *questions = NULL;
    SubjectList *currentList = list, *holderList = NULL;
    const char *holder = NULL;
    char input[LINE_BUFFER_SIZE];
    Property *curProp = NULL;
    Question *currentQuestion = NULL;

    if (list->count() == 0) {
        holder = list->sourceFile();

        if (holder) {
            cout << "There are no entries in \"" << holder << "\".\n\n";
        } else {
            cout << "No entries to search.\n\n";
        }
        
        return;
    } else if (list->count() == 1) { // Dude, c'mon...
        displayList(list, true);
        cout << endl;
        return;
    }

    questions = new List<Question>();
    
    while ((currentQuestion = generateNextQuestion(currentList, questions, true)) != NULL) {
        questions->append(currentQuestion);
        currentQuestion->release();
        
        lastCount = currentList->count();

        if (questionIsRelevant(currentList, currentQuestion)) {
            if (currentQuestion->isBoolean()) {
                cout << "Is the subject " << currentQuestion->getQuestion() << "?  ";
                currentQuestion->setBooleanAnswer(getBoolean(cin));

                curProp = new Property(copyOfString(currentQuestion->getQuestion()), NULL);
                holderList = findAllSubjectByProperty(currentList, curProp, currentQuestion->getBooleanAnswer());
            } else {
                cout << currentQuestion->getQuestion() << ":  ";
                safeGetline(cin, input, LINE_BUFFER_SIZE);

                currentQuestion->setStringAnswer(copyOfString(input));
                
                if (caseInsensitiveStringCompare(input, "remove") != 0) {
                    curProp = new Property(copyOfString(currentQuestion->getQuestion()), copyOfString(input));
                    holderList = findAllSubjectByProperty(currentList, curProp, true);
                } else {
                    curProp = new Property(copyOfString(currentQuestion->getQuestion()), NULL);
                    holderList = findAllSubjectByProperty(currentList, curProp, false);
                }
            }

            curProp->release();
            
            if (currentList != list) {
                delete currentList;
            }

            currentList = holderList;
            
            if (currentList->count() != lastCount) {
                cout << "This reduces the list of possible matches to " << currentList->count() << " subject.  ";

                if (currentList->count() > 0) {
                    if (currentList->count() <= VERBOSE_LIST_THRESHOLD) {
                        cout << "Listed below.\n\n";
                        displayList(currentList, true);
                    } else if (currentList->count() <= BRIEF_LIST_THRESHOLD) {
                        cout << "Listed below.\n\n";
                        displayList(currentList, false);
                    } else {
                        cout << "Too many to show.\n\n";
                    }

                    if (currentList->count() == 1) {
                        cout << "Is this the subject you were looking for?  ";

                        if (!getBoolean(cin)) {
                            cout << "Would you like to add a new subject?  ";

                            if (getBoolean(cin)) {
                                addSubjectToList(list, questions);
                            }
                        } else {
                            cout << endl;
                        }

                        delete questions;

                        if (currentList != list) {
                            delete currentList;
                        }
                        
                        return;
                    }                        
                } else {
                    cout << "No match found.\n\nWould you like to add a new subject?  ";

                    if (getBoolean(cin)) {
                        addSubjectToList(list, questions);
                    } else {
                        cout << endl;
                    }

                    delete questions;

                    if (currentList != list) {
                        delete currentList;
                    }
                    
                    return;
                }
            }
        }
    }

    cout << "No more questions to ask!  All remaining items appear to be identical.\n";
    cout << "Is this the subject you were looking for?  ";
    
    if (!getBoolean(cin)) {
        cout << "Would you like to add a new subject?  ";

        if (!getBoolean(cin)) {
            cout << "Would you like to add a new subject?  ";

            if (getBoolean(cin)) {
                addSubjectToList(list, questions);
            }
        } else {
            cout << endl;
        }

        delete questions;

        if (currentList != list) {
            delete currentList;
        }
        
        return;
    }

    cout << endl;
    delete questions;

    if (currentList != list) {
        delete currentList;
    }
}
