/*  Student Name: Wade Tregaskis
    Student Number: 02557793 */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "subjectlib.h"


/* generateNextQuestion(List *list, List *pastQuestions, bool tryToBeRandom)
     Returns a QuestionItem instance containing the next question to be asked of
     the subjects in 'list'.  If 'pastQuestions' is not NULL, it will ensure it
     does not return any question that is present in 'pastQuestions'.  If
     'tryToBeRandom' is true, this function will obviously try to pick the
     question somewhat randomly (i.e. such that the order is not predictable),
     but makes no guarantees.
       If there are no questions that can be asked of 'list' that don't already
     appear in 'pastQuestions', returns NULL. */

Question* generateNextQuestion(SubjectList *list, List<Question> *pastQuestions, bool tryToBeRandom) {
    Cursor<Subject> *subjectCursor = NULL;
    Cursor<Property> *propertyCursor = NULL;
    Subject *currentSubject = NULL;
    Property *currentProperty = NULL;
    Question *result = NULL, *fallback = NULL;
    
    if (list) {
        subjectCursor = new Cursor<Subject>(list);

        while ((currentSubject = subjectCursor->getCurrent()) != NULL) {
            propertyCursor = new Cursor<Property>(currentSubject->propertyList());

            while ((currentProperty = propertyCursor->getCurrent()) != NULL) {
                if (!questionIsInList(currentProperty->getName(), pastQuestions)) {
                    result = new Question(currentProperty);
                    
                    if (fallback) {
                        fallback->release();
                    }
                    
                    if (!tryToBeRandom || ((int)(list->count() * (rand() / (RAND_MAX + 1.0))) == 0)) {
                        delete subjectCursor;
                        delete propertyCursor;

                        return result;
                    } else {
                        fallback = result;
                    }
                }

                propertyCursor->moveToNext();
            }

            delete propertyCursor;
            
            subjectCursor->moveToNext();
        }

        delete subjectCursor;
    }
	 
    return fallback;
}

/* questionIsInList(const char *question, List *questionList)
     Returns true if the 'question' appears in 'list'. */

bool questionIsInList(const char *question, List<Question> *questionList) {
    Cursor<Question> *c = NULL;
    Question *current = NULL;
    bool found = false;
    
    if (questionList) {
        c = new Cursor<Question>(questionList);

        while (!found && (current = c->getCurrent()) != NULL) {
            if (caseInsensitiveStringCompare(question, current->getQuestion()) == 0) {
                found = true;
            }

            c->moveToNext();
        }

        delete c;
    }

    return found;
}

/* questionIsRelevant(SubjectList *list, QuestionItem *item)
     Returns true if the question of 'item' is relevant to 'list'.  This is in
     terms of whether asking it makes any sense for that list - it is *not*
     based necessarily on whether the properties of 'item' appear in 'list', or
     similar.  e.g. if all items in 'list' have the property of 'item', this
     will return false - asking that question won't help you gain any further
     info on what subject you are looking for. */

bool questionIsRelevant(SubjectList *list, Question *question) {
    Cursor<Subject> *listCursor = NULL;
    Cursor<Property> *propertyCursor = NULL;
    int itemCount = 0;
    Subject *currentSubject = NULL;
    Property *currentProperty = NULL, *lastProperty = NULL;
    bool foundMatch, booleanMode, gotResult = false, result = false, haveHadTrue = false, haveHadFalse = false;
    
    if (!list || !question || (list->count() < 1)) {
        return false; // one could argue :)
    }

    booleanMode = question->isBoolean();
    listCursor = new Cursor<Subject>(list);

    while (!gotResult && ((currentSubject = listCursor->getCurrent()) != NULL)) {
        ++itemCount;

        foundMatch = false;
        propertyCursor = new Cursor<Property>(currentSubject->propertyList());

        while (!gotResult && ((currentProperty = propertyCursor->getCurrent()) != NULL)) {
            if ((currentProperty->isBoolean() == booleanMode) && (caseInsensitiveStringCompare(currentProperty->getName(), question->getQuestion()) == 0)) {
                foundMatch = true;

                if (booleanMode) {
                    if (haveHadFalse) {
                        gotResult = true;
                        result = true;
                    } else {
                        haveHadTrue = true;
                    }
                } else {
                    if (lastProperty && (strcmp(lastProperty->getValue(), currentProperty->getValue()) != 0)) {
                        gotResult = true;
                        result = true;
                    } else {
                        lastProperty = currentProperty;
                    }
                }
            }

            propertyCursor->moveToNext();
        }

        delete propertyCursor;

        if (!gotResult && !foundMatch && booleanMode) {
            if (haveHadTrue) {
                gotResult = true;
                result = true;
            } else {
                haveHadFalse = true;
            }
        }

        listCursor->moveToNext();
    }

    delete listCursor;

    return result;
}

/* findAllSubjectByProperty(SubjectList *list, Property *prop, bool with)
     If 'with' is true, returns a list of all subject which have the property 'prop',
     from 'list'.  If 'with' is false, returns a list of all subject which do not have
     the property 'prop'.
        For boolean properties, the presence of the property determines if a subject
     is matched.  For string properties, the value of the property must the same as that
     of 'prop' for it to be considered a match. */

SubjectList* findAllSubjectByProperty(SubjectList *list, Property *prop, bool with) {
    SubjectList *result = NULL;
    Cursor<Subject> *listCursor = NULL;
    Cursor<Property> *propertyCursor = NULL;
    Subject *currentSubject = NULL;
    Property *currentProperty = NULL;
    bool foundProperty;

    if (list && prop) {
        result = new SubjectList();
        listCursor = new Cursor<Subject>(list);

        while ((currentSubject = listCursor->getCurrent()) != NULL) {
            propertyCursor = new Cursor<Property>(currentSubject->propertyList());
            foundProperty = false;
            
            while (!foundProperty && (currentProperty = propertyCursor->getCurrent()) != NULL) {
                if (caseInsensitiveStringCompare(currentProperty->getName(), prop->getName()) == 0) {
                    if (prop->isBoolean()) {
                        foundProperty = true;
                    } else if ((prop->getValue() == NULL) || (caseInsensitiveStringCompare(currentProperty->getValue(), prop->getValue()) == 0)) {
                        foundProperty = true;
                    }
                }

                propertyCursor->moveToNext();
            }

            delete propertyCursor;

            if (with == foundProperty) {
                result->append(currentSubject);
            }

            listCursor->moveToNext();
        }

        delete listCursor;
    }

    return result;
}

// -----------------------------
// IMPLEMENTATION:  SHAREDOBJECT
// -----------------------------

SharedObject::SharedObject(void) {
    retainCount = 1;
}

SharedObject::~SharedObject(void) {
    cleanup();
}

void SharedObject::cleanup(void) {
    if (retainCount != 0) {
        cout << "DEBUG Warning:  SharedObject cleanup() called for " << this << " [retain count == " << retainCount << "]\n";
    }
}

void SharedObject::retain(void) {
    ++retainCount;
}

void SharedObject::release(void) {
    --retainCount;

    if (retainCount < 1) {
        cleanup();
        delete this;
    }
}

int SharedObject::getRetainCount(void) const {
    return retainCount;
}

void SharedObject::output(ostream& out) const {
    out << "SharedObject &h" << this << ", retainCount == " << retainCount << endl;
}

// ------------------------
// IMPLEMENTATION:  SPECIES
// ------------------------

Subject::Subject(char *newName):SharedObject() {
    name = newName;
    property = new List<Property>();
}

Subject::Subject(const Subject &original) {
    if (original.name) {
        name = copyOfString(original.name);
    } else {
        name = NULL;
    }

    property = new List<Property>(*(original.property));
}

Subject::~Subject(void) {
    cleanup();
}

void Subject::cleanup(void) {
    int retCount;

    retCount = getRetainCount();
    
    if (retCount > 0) {
        cout << "DEBUG Warning:  Subject " << this << " cleaned up while retain count == " << retCount << endl;
    }
        
    if (name) {
        delete[] name;
    }

    if (property) {
        delete property;
    }
}

const char* Subject::getName(void) const {
    return name;
}

void Subject::setName(char *newName) {
    if (name) {
        delete[] name;
    }

    name = newName;
}

List<Property>* Subject::propertyList(void) const {
    return property;
}

void Subject::output(ostream& out) const {
    out << name << endl << *property;
}
    
// ----------------------------
// IMPLEMENTATION:  SPECIESLIST
// ----------------------------

SubjectList::SubjectList(char *sourceFile):List<Subject>() {
    openedFrom = sourceFile;
    
    if (sourceFile) {
        appendContentsOfFile(sourceFile);
    }
}

SubjectList::SubjectList(const SubjectList &original):List<Subject>() {
    if (original.openedFrom) {
        openedFrom = copyOfString(original.openedFrom);
    }

    append(&original);
}

SubjectList::~SubjectList(void) {
    if (openedFrom) {
        delete[] openedFrom;
    }
}

bool SubjectList::appendContentsOfFile(char *fileName) {
    bool newEntry = true;
    ifstream fin;
    char curLine[LINE_BUFFER_SIZE];
    Subject *tempItem = NULL;
    Property *tempProperty = NULL;
    List<Property> *properties = NULL;

    if (fileName) {
        fin.open(fileName);

        if (fin.fail()) {
            return false;
        }

        while(!fin.eof()) {
            safeGetline(fin, curLine, LINE_BUFFER_SIZE);

            if (strlen(curLine) == 0) { // Empty line
                if (!newEntry) {
                    append(tempItem);
                    tempItem->release();
                }

                newEntry = true;
            } else {
                if (newEntry) {
                    tempItem = new Subject(copyOfString(curLine));
                    properties = tempItem->propertyList();
                    
                    newEntry = false;
                } else {
                    tempProperty = new Property(curLine);
                    properties->append(tempProperty);
                    tempProperty->release();
                }
            }
        }

        fin.close();
    }
    
    return true;
}

bool SubjectList::saveToFile(char *fileName) const {
    ofstream out;

    if (fileName) {
        out.open(fileName);
    } else {
        out.open(openedFrom);
    }

    if (out.fail()) {
        return false;
    } else {
        out << *this;
        out.close();

        return true;
    }
}

const char* SubjectList::sourceFile(void) const {
    return openedFrom;
}

// -------------------------
// IMPLEMENTATION:  PROPERTY
// -------------------------

Property::Property(const char *newName):SharedObject() {
    int start, length = 1, i = 1;
    
    if (newName && (newName[0] == '#')) { // #Name:value
        boolean = false;

        while((newName[i] != ':') && (newName[i] != '\0')) {
            ++i;
            ++length;
        }

        name = new char[length];
        name[length - 1] = '\0';
        strncpy(name, newName + 1, length - 1);

        if (newName[i] != '\0') {
            start = i = length + 1;
            length = 1;

            while(newName[i] != '\0') {
                ++i;
                ++length;
            }

            value = new char[length];
            value[length - 1] = '\0';
            strncpy(value, newName + start, length - 1);
        } else {
            value = NULL;
        }
    } else { // Boolean property
        boolean = true;
        name = copyOfString(newName);
        value = NULL;
    }
}

Property::Property(char *newName, char *newValue):SharedObject() {
    name = newName;
    value = newValue;

    if (newValue) {
        boolean = false;
    } else {
        boolean = true;
    }
}

Property::Property(const Property &original):SharedObject() {
    name = copyOfString(original.name);
    value = copyOfString(original.value);
    boolean = original.boolean;
}

Property::~Property(void) {
    cleanup();
}

void Property::cleanup(void) {
    if (getRetainCount() > 0) {
        cout << "DEBUG Warning:  Property " << this << " cleaned up while retain count == " << getRetainCount() << endl;
    }
    
    if (name) {
        delete[] name;
    }

    if (value) {
        delete[] value;
    }
}

const char* Property::getName(void) const {
    return name;
}

const char* Property::getValue(void) const {
    return value;
}

void Property::setName(char *newName) {
    if (name) {
        delete[] name;
    }

    name = newName;
}

void Property::setValue(char *newValue) {
    boolean = false;

    if (value) {
        delete[] value;
    }

    value = newValue;
}

bool Property::isBoolean(void) const {
    return boolean;
}

void Property::output(ostream& out) const {
    if (boolean) {
        out << name;
    } else {
        out << '#' << name << ':' << value;
    }
}

// -------------------------
// IMPLEMENTATION:  QUESTION
// -------------------------

Question::Question(char *theQuestion, bool value):SharedObject() {
    question = theQuestion;
    booleanValue = value;

    boolean = true;
}

Question::Question(char *theQuestion, char *theAnswer):SharedObject() {
    question = theQuestion;
    answer = theAnswer;

    boolean = false;
}

Question::Question(const Question &original):SharedObject() {
    question = copyOfString(original.question);
    answer = copyOfString(original.answer);
    boolean = original.boolean;
    booleanValue = original.booleanValue;
}

Question::Question(const Property *basedOn):SharedObject() {
    answer = NULL;
        
    if (basedOn) {
        question = copyOfString(basedOn->getName());
        boolean = basedOn->isBoolean();
    } else {
        question = NULL;
        boolean = false;
    }
}

Question::~Question(void) {
    cleanup();
}

void Question::cleanup(void) {
    if (getRetainCount() > 0) {
        cout << "DEBUG Warning:  Question " << this << " cleaned up while retain count == " << getRetainCount() << endl;
    }
    
    if (question) {
        delete[] question;
    }

    if (answer) {
        delete[] answer;
    }
}

const char* Question::getQuestion(void) const {
    return question;
}

const char* Question::getStringAnswer(void) const {
    return answer;
}

bool Question::getBooleanAnswer(void) const {
    return booleanValue;
}

void Question::setQuestion(char *newQuestion) {
    if (question) {
        delete[] question;
    }

    question = newQuestion;
}

void Question::setStringAnswer(char *newAnswer) {
    if (answer) {
        delete[] answer;
    }

    answer = newAnswer;
}

void Question::setBooleanAnswer(const bool newAnswer) {
    booleanValue = newAnswer;
}

bool Question::isBoolean(void) const {
    return boolean;
}

void Question::output(ostream& out) const {
    if (boolean) {
        out << '?' << question << ": " << boolToString(booleanValue);
    } else {
        out << '?' << question << ": " << answer;
    }
}

// ---------
// OPERATORS
// ---------

// Mega evil slow; don't use unless you really really must.
// Returns true if both lists contain the exact same items, in any order
template<class T>
bool operator==(const List<T>& left, const List<T>& right) {
    Cursor<T> *a = NULL, *b = NULL;
    T *current = NULL;

    if (left.count() != right.count()) {
        return false;
    }

    a = new Cursor<T>(&left);
    b = new Cursor<T>(&right);

    while ((current = a->getCurrent()) != NULL) {
        if (!(b->moveTo(current)) || (current != b->getCurrent())) {
            delete a;
            delete b;
            
            return false;
        }
    }

    delete a;
    delete b;
    
    return true;
}

template<class T>
inline bool operator!=(const List<T>& left, const List<T>& right) {
    return !(left == right);
}

// Mega evil slow; don't use unless you really really must.
// Returns true if both lists contain the exact same items, in any order
bool operator==(const SubjectList& left, const SubjectList& right) {
    Cursor<Subject> *a = NULL, *b = NULL;
    Subject *current = NULL;

    if ((left.count() != right.count()) || (strcmp(right.sourceFile(), left.sourceFile()) != 0)) {
        return false;
    }

    a = new Cursor<Subject>(&left);
    b = new Cursor<Subject>(&right);

    while ((current = a->getCurrent()) != NULL) {
        if ((!(b->moveTo(current))) || (*current != *(b->getCurrent()))) {
            delete a;
            delete b;
            
            return false;
        }
    }

    delete a;
    delete b;
    
    return true;
}

inline bool operator!=(const SubjectList& left, const SubjectList& right) {
    return !(left == right);
}

bool operator==(const Subject& left, const Subject& right) {
    List<Property> *a = NULL, *b = NULL;
    
    if (strcmp(left.getName(), right.getName()) != 0) {
        return false;
    }

    a = left.propertyList();
    b = right.propertyList();

    if (*a != *b) {
        return false;
    }

    return true;
}

inline bool operator!=(const Subject& left, const Subject& right) {
    return !(left == right);
}

bool operator==(const Property& left, const Property& right) {
    if (strcmp(left.getName(), right.getName()) != 0) {
        return false;
    }

    if (left.isBoolean() != right.isBoolean()) {
        return false;
    }

    if (!left.isBoolean()) {
        if (strcmp(left.getValue(), right.getValue()) != 0) {
            return false;
        }
    }

    return true;
}

inline bool operator!=(const Property& left, const Property& right) {
    return !(left == right);
}

bool operator==(const Question& left, const Question& right) {
    if (strcmp(left.getQuestion(), right.getQuestion()) != 0) {
        return false;
    }

    if (left.isBoolean() != right.isBoolean()) {
        return false;
    }

    if (left.isBoolean()) {
        if (left.getBooleanAnswer() != right.getBooleanAnswer()) {
            return false;
        }
    } else {
        if (strcmp(left.getStringAnswer(), right.getStringAnswer()) != 0) {
            return false;
        }
    }

    return true;
}

inline bool operator!=(const Question& left, const Question& right) {
    return !(left == right);
}

template<class T>
ostream& operator<<(ostream& out, const List<T>& list) {
    Cursor<T> *c = NULL;
    T *current = NULL;
    
    c = new Cursor<T>(&list);

    while ((current = c->getCurrent()) != NULL) {
        out << *current << endl;
        c->moveToNext();
    }

    delete c;

    return out;
}

ostream& operator<<(ostream &out, const SharedObject& ob) {
    ob.output(out);
    return out;
}
