/*  Student Name: Wade Tregaskis
    Student Number: 02557793 */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef SUBJECTLIB_H
#define SUBJECTLIB_H

// -INCLUDES-

//#include <ctime>
#include <cstdlib>
#include <cstring>
#include <iostream> // For displaying DEBUG messages
#include <fstream>
#include "ioutils.h"
#include "utils.h"
using namespace std;


// -DEFINES-

#define SUBJECTLIB_VERSION 2.26


// -CLASSES-

class SharedObject;
template<class T> class List; // Implemented 100% inline
class SubjectList;
template<class T> class Cursor; // Implemented 100% inline
template<class T> class EditCursor; // Implemented 100% inline
class Subject;
class Property;
class Question;
template<class T> class ListItem; // Implemented 100% inline

template<class T> class List {
protected:
    int numberOfEntries;
    ListItem<T> *first, *last;
    
public:
    List(void) { numberOfEntries = 0; first = last = NULL; }
    List(const List<T> &original) { numberOfEntries = 0; first = last = NULL; append(&original); }
    ~List(void) {
        ListItem<T> *temp = first, *next;

        while (temp) {
            next = temp->getNext();
            delete temp;
            temp = next;
        }
    }

    int count(void) const { return numberOfEntries; }    

    void append(T *data) {
        ListItem<T> *newItem;

        if (data) {
            newItem = new ListItem<T>(data, this);

            if (last) { // At least one entry already
                newItem->setPrevious(last);
                last->setNext(newItem);
                last = newItem;
            } else  { // No entries yet in list
                first = newItem;
                last = newItem;
            }

            increaseCount();
        }
    }

    void append(const List<T> *list) {
        ListItem<T> *currentItem;

        currentItem = list->getFirst();

        while (currentItem) {
            append(currentItem->getTarget());
            currentItem = currentItem->getNext();
        }
    }
    
    ListItem<T>* getFirst(void) const { return first; }
    ListItem<T>* getLast(void) const { return last; }

    void setFirst(ListItem<T> *item) { first = item; }
    void setLast(ListItem<T> *item) { last = item; }
    
    void reduceCount(int by = 1) { numberOfEntries -= by; }
    void increaseCount(int by = 1) { numberOfEntries += by; }
};

class SubjectList: public List<Subject> {
private:
    char *openedFrom;
    
public:
    SubjectList(char *sourceFile = NULL);
    SubjectList(const SubjectList &original);
    ~SubjectList();
    
    bool appendContentsOfFile(char *fileName);
    bool saveToFile(char *fileName = NULL) const;

    const char* sourceFile(void) const;
};

template<class T> class Cursor {
protected:
    ListItem<T> *current;
    const List<T> *owner;
    
public:
    Cursor(const List<T>* newOwner = NULL) { setOwner(newOwner); }
    Cursor(const Cursor<T> &original) { current = original.current; owner = original.owner; }    

    T* getAt(int index) const {
        ListItem<T> *temp = NULL;
        int count;

        if (owner) {
            if (index < 0) {
                return NULL;
            }

            count = owner->count() - 1;

            if (index > count) {
                return NULL;
            }

            // The following branch attempts to improve speed by starting from the closest end
            if (index < (count / 2)) {
                count = 0;
                temp = owner->getFirst();

                while (temp && (index != count)) {
                    ++count;
                    temp = temp->getNext();
                }
            } else {
                temp = owner->getLast();

                while (temp && (index != count)) {
                    --count;
                    temp = temp->getPrevious();
                }
            }
        }

        return temp->getTarget();
    }    

    T* getNext(void) const {
        ListItem<T> *temp;

        if (current) {
            temp = current->getNext();

            if (temp) {
                return temp->getTarget();
            }
        }

        return NULL;
    }

    T* getPrevious(void) const {
        ListItem<T> *temp;

        if (current) {
            temp = current->getPrevious();

            if (temp) {
                return temp->getTarget();
            }
        }

        return NULL;
    }

    T* getCurrent(void) const {
        if (current) {
            return current->getTarget();
        } else {
            return NULL;
        }
    }
    
    bool jumpForward(int count) {
        while ((count > 0) && current) {
            --count;
            current = current->getNext();
        }

        return (current != NULL);
    }

    bool jumpBack(int count) {
        while ((count > 0) && current) {
            --count;
            current = current->getPrevious();
        }

        return (current != NULL);
    }    

    void moveToNext(void) {
        if (current) {
            current = current->getNext();
        }
    }

    void moveToPrevious(void) {
        if (current) {
            current = current->getPrevious();
        }
    }

    void moveToStart(void) {
        if (owner) {
            current = owner->getFirst();
        } else {
            current = NULL;
        }
    }
    
    void moveToEnd(void) {
        if (owner) {
            current = owner->getLast();
        } else {
            current = NULL;
        }
    }

    bool moveTo(const T *target) {
        if (owner) {
            current = owner->getFirst();

            while (current && (current->getTarget() != target)) {
                current = current->getNext();
            }

            return (current != NULL);
        } else {
            return false;
        }
    }    

    const List<T>* getOwner(void) const { return owner; }
    void setOwner(const List<T> *newOwner) { owner = newOwner; moveToStart(); }
};

template<class T> class EditCursor: private Cursor<T> {
protected:
    List<T> *owner;
    
public:
    EditCursor(List<T>* newOwner) { setOwner(newOwner); }
    
    void remove(void) {
        ListItem<T> *temp = NULL;

        if (this->current && owner) {
            if (this->current->getNext() == NULL) {
                temp = this->current->getPrevious();
            } else {
                temp = this->current->getNext();
            }

            delete this->current;

            this->current = temp;
        }
    }    

    void insertAfter(T* newEntry) const {
        ListItem<T> *temp = NULL, *newItem = NULL;

        if (owner && newEntry && this->current) {
            newItem = new ListItem<T>(newEntry, owner);

            temp = this->current->getNext();

            newItem->setNext(temp);
            newItem->setPrevious(this->current);

            temp->setPrevious(newItem);
            this->current->setNext(newItem);

            owner->increaseCount();
        }
    }

    void insertBefore(T* newEntry) const {
        ListItem<T> *temp = NULL, *newItem = NULL;

        if (owner && newEntry && this->current) {
            newItem = new ListItem<T>(newEntry, owner);

            temp = this->current->getPrevious();

            newItem->setNext(this->current);
            newItem->setPrevious(temp);

            temp->setNext(newItem);
            this->current->setPrevious(newItem);

            owner->increaseCount();
        }
    }

    void replaceWith(T* newEntry) {
        ListItem<T> *old = NULL, *temp = NULL, *newItem = NULL;

        if (owner && newEntry && this->current) {
            newItem = new ListItem<T>(newEntry, owner);

            old = this->current;
            this->current = newItem;

            if (old == owner->getFirst()) {
                owner->setFirst(newItem);
            }

            if (old == owner->getLast()) {
                owner->setLast(newItem);
            }

            temp = old->getPrevious();

            if (temp) {
                temp->setNext(newItem);
            }

            temp = old->getNext();

            if (temp) {
                temp->setPrevious(newItem);
            }

            old->setNext(NULL);
            old->setPrevious(NULL);

            delete old;
        }
    }    

    List<T>* getOwner(void) const { return owner; }
    void setOwner(List<T> *newOwner) { owner = newOwner; this->moveToStart(); }
};

class SharedObject {
private:
    int retainCount;
    
public:
    SharedObject(void);
    ~SharedObject(void);
    virtual void cleanup(void);
    
    void retain(void);
    void release(void);

    int getRetainCount(void) const;

    virtual void output(ostream& out) const;
};

class Subject: public SharedObject {
private:
    char *name;
    List<Property> *property;
    
public:
    Subject(char *newName = NULL);
    Subject(const Subject &original);
    ~Subject(void);
    void cleanup(void);
    
    const char* getName(void) const;
    void setName(char *newName);

    List<Property>* propertyList(void) const;

    virtual void output(ostream& out) const;
};

class Property: public SharedObject {
private:
    char *name, *value;
    bool boolean;
    
public:
    Property(const char *newName = NULL);
    Property(char *newName, char *newValue);
    Property(const Property &original);
    ~Property(void);
    void cleanup(void);

    const char* getName(void) const;
    const char* getValue(void) const;
    
    void setName(char *newName);
    void setValue(char *newValue);
    
    bool isBoolean(void) const;

    virtual void output(ostream& out) const;
};

class Question: public SharedObject {
private:
    char *question, *answer;
    bool boolean, booleanValue;
    
public:
    Question(char *theQuestion = NULL, bool value = true);
    Question(char *theQuestion, char *theAnswer);
    Question(const Question &original);
    Question(const Property *basedOn);
    ~Question(void);
    void cleanup(void);

    const char* getQuestion(void) const;
    const char* getStringAnswer(void) const;
    bool getBooleanAnswer(void) const;

    void setQuestion(char *newQuestion);
    void setStringAnswer(char *newAnswer);
    void setBooleanAnswer(const bool newAnswer);

    bool isBoolean(void) const;

    virtual void output(ostream& out) const;
};

template<class T> class ListItem {
private:
    ListItem<T> *previous, *next;
    List<T> *owner;
    T *target;
    
public:
    ListItem(T *theTarget = NULL, List<T> *theOwner = NULL) {
        previous = next = NULL;

        target = theTarget;
        owner = theOwner;

        if (target) {
            target->retain();
        }
    }

    ListItem(const ListItem<T> &original) {
        next = previous = NULL;
        owner = original.owner;
        target = original.target;

        if (target) {
            target->retain();
        }
    }

    ~ListItem(void) {
        if (target) {
            target->release();
        }

        if (owner) {
            owner->reduceCount();

            if (this == owner->getFirst()) {
                owner->setFirst(next);
            }

            if (this == owner->getLast()) {
                owner->setLast(previous);
            }
        }

        if (next) {
            next->setPrevious(previous);
        }

        if (previous) {
            previous->setNext(next);
        }
    }    

    ListItem<T>* getNext(void) const { return next; }
    ListItem<T>* getPrevious(void) const { return previous; }
    void setNext(ListItem<T> *target) { next = target; }
    void setPrevious(ListItem<T> *target) { previous = target; }

    T* getTarget(void) const { return target; }
    List<T>* getOwner(void) const { return owner; }
};


// -FUNCTIONS-

Question* generateNextQuestion(SubjectList *list, List<Question> *pastQuestions, bool tryToBeRandom);
bool questionIsInList(const char *question, List<Question> *questionList);
bool questionIsRelevant(SubjectList *list, Question *question);
SubjectList* findAllSubjectByProperty(SubjectList *list, Property *prop, bool with);

template<class T> bool operator==(const List<T>& left, const List<T>& right);
template<class T> bool operator!=(const List<T>& left, const List<T>& right);
bool operator==(const SubjectList& left, const SubjectList& right);
bool operator!=(const SubjectList& left, const SubjectList& right);
bool operator==(const Subject& left, const Subject& right);
bool operator!=(const Subject& left, const Subject& right);
bool operator==(const Property& left, const Property& right);
bool operator!=(const Property& left, const Property& right);
bool operator==(const Question& left, const Question& right);
bool operator!=(const Question& left, const Question& right);

template<class T> ostream& operator<<(ostream& out, const List<T>& list);
ostream& operator<<(ostream& out, const SharedObject& ob);

#endif
