/*  Student Name: Wade Tregaskis
    Student Number: 02557793 */

//  Copyright (c) 2006, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "utils.h"


/* removeOutsideWhitespace(char **data)
     Takes the 'data' and removes any whitespace before or after any text in it.
     Note that 'data' is a pointer to a pointer to a string (phew!), so for a
     normal *char pointer, you'll need to prefix & to the pointer name when
     passing to this function. */

void removeOutsideWhitespace(char **data) {
    int textStart, textEnd = 0, i, length, newLength;
    char *result, *realData;

    realData = *data;
    length = strlen(realData);

    textStart = length;

    for (i = 0; i < length; ++i) {
        if (!isspace(realData[i])) {
            textStart = i;
            break;
        }
    }

    for (i = length; i > 0; --i) {
        if (!isspace(realData[i - 1])) {
            textEnd = i;
            break;
        }
    }

    newLength = (textEnd - textStart);

    if (newLength > 0) { // malloc doesn't like negative values
        result = new char[newLength + 1];

        for (i = 0; i < newLength; ++i) {
            result[i] = realData[textStart + i];
        }

        result[newLength] = '\0';
    } else { // Indicates the original string was pure whitespace
        result = new char[1];
        result[0] = '\0';
    }

    delete[] realData;
    *data = result;
}

/* copyOfString(const char *data)
     Returns a copy of the string 'data'. */

char* copyOfString(const char *data) {
    char *result = NULL;
    int length;

    if (data) {
        length = strlen(data) + 1;

        result = new char[length];
        strncpy(result, data, length);
    }
    
    return result;
}

/* boolToString(bool value)
     Converts the boolean passed to a string value.  Defaults to true = "Yes",
     false = "No". */

const char* boolToString(bool value) {
    static char strYes[] = "Yes", strNo[] = "No";

    if (value) {
        return strYes;
    } else {
        return strNo;
    }
}

/* boolFromString(const char *data)
     Converts the string passed to a boolean value, using a wide variety of
     techniques to hopefully maximize compatability.  Returns false by default. */

bool boolFromString(const char *data) {
    if (strlen(data)) {
        switch (toupper(data[0])) {
        case 'Y': // Yes
            return true;
        case 'N': // No/Negative
            return false;
        case '1': // 1
            return true;
        case '0': // 0
            return false;
        case 'T': // True
            return true;
        case 'F': // False
            return false;
        case '+': // +
            return true;
        case '-': // -
            return false;
        }
    }

    return false;
}

/* isBoolean(const char *data)
     Returns whether or not the passed string contains a boolean value [so far as
     it can tell]. */

bool isBoolean(const char *data) {
    char smallTemp[2] = {'\0', '\0'};

    if (data) {
        if (strlen(data) == 1) {
            smallTemp[0] = toupper(data[0]);

            return (strpbrk(smallTemp, "YNTF10+-") != NULL);
        } else {
            return (caseInsensitiveStringCompare(data, "yes") == 0 ||
                    caseInsensitiveStringCompare(data, "no") == 0 ||
                    caseInsensitiveStringCompare(data, "true") == 0 ||
                    caseInsensitiveStringCompare(data, "false") == 0);
        }
    } else {
        return false;
    }
}

/* caseInsensitiveStringCompare(const char *a, const char *b)
     Compares the two strings 'a' and 'b' without regard to their case.  Returns
     -1 if 'a' is less than 'b', 1 if 'b' is less than 'a', and 0 if they are
     identical - i.e. the same as strcmp(). */

int caseInsensitiveStringCompare(const char *a, const char *b) {
    int offset, end, lengthA, lengthB, charA, charB;

    if (a == b) {
        return 0;
    }
    
    lengthA = strlen(a);
    lengthB = strlen(b);
    
    end = (lengthA < lengthB) ? lengthA : lengthB;

    for(offset = 0; offset < end; ++offset) {
        charA = toupper(a[offset]);
        charB = toupper(b[offset]);

        if (charA != charB) {
            return (charA < charB) ? -1 : 1;
        }
    }

    if (lengthA == lengthB) {
        return 0;
    } else {
        return (lengthA < lengthB) ? -1 : 1;
    }
}
